#pragma once
#include <stack>
#include <unordered_map>
#include <string>
#include <exception>
#include <ostream>
#include <iostream>
#include <type_traits>
#include "ast/type_info.hpp"
#include "llvm/IR/Value.h"

namespace symbol_table {
    class Symb {
       private:
           std::string name;

       public:
           Symb() {this->name = "";}
           Symb(std::string name) { this->name = name;}
           std::string get_name() const {return name;}
           operator std::string() const {return name;}
    };

    /*
    classes used to hold information associated with Symbs
    (which are stored in SymbolTables)
    */
    class SymbolInfo: public Symb {
        public:
            type_info::Kind kind;
            type_info::Type type;
            bool valid;

            // Quantum-specific info
            bool isQuantum;
            std::pair<unsigned int, unsigned int> address;

            SymbolInfo();

            // Copy Constructor
            SymbolInfo(const SymbolInfo& other);

            SymbolInfo(std::string id, type_info::Kind kind, type_info::Type type);

            SymbolInfo(std::string id, type_info::KindValue kind, type_info::TypeValue type);

            /*
            SymbolInfo(std::string id, quant_c_ast::QuantumContext quantum_context, type_info::KindValue kind, type_info::TypeValue type);
            */

            operator std::string() const;
    };

    class SymbolValue: public Symb {
        public:
            llvm::Value* codegenValue;

            SymbolValue();

            // Copy Constructor
            SymbolValue(const SymbolValue& other);

            SymbolValue(std::string id, llvm::Value* value);

            operator std::string() const;
    };

    /*
     A block-structured symbol table class which stores symbol objects into a scopes.
     */
    template <class Symbol>
    class SymbolTable {

        /*
        Data structure used to store scoped symbols.
        A scope is implemented as an std::unordered_map<std::string, Symb>, where a symbol's name
        acts as the key to the actual symbol object in that scope.
        */
        private:
            std::stack<std::unordered_map<std::string, Symbol>> blockSymbolStack;

        public:
            SymbolTable();

            /*
            Creates a new empty scope in this SymbolTable and sets it to be
            the current scope.
            */
            void openScope();

            /*
            Removes the current scope from this SymbolTable.
            If the SymbolTable is empty, an EmptySTException is thrown.
            */
            void closeScope();

            /*
            Inserts a symbol into the current scope of this SymbolTable.
            If the SymbolTable is empty, an EmptySTException is thrown.
            If the current scope already has a symbol with the same name as the
            provided symbol, a DuplicateException is thrown.
            */
            void insert(const Symbol sym);

            /*
            Locates a symbol by the provided name in the current scope.
            Returns the symbol object if its name is found, otherwise null is returned.
            */
            Symbol& localLookup(const std::string name);

            /*
            Locates a symbol by the provided name in any of the scopes present
            in this SymbolTable, starting with the current scope and working
            backwards to prior scopes.
            Returns the symbol object if its name is found, otherwise null is returned.
            */
            Symbol& globalLookup(const std::string name);

            /*
            Returns a string representation of this SymbolTable.
            Information included in the string includes all open scopes
            (in order from current to first)
            and the symbols contained within each scope.
            */
            operator std::string() const;

            // Write the contents of this SymbolTable into the provided stream.
            void dump(std::ostream &ps);

            // Return number of scopes open in the symbol table.
            int scopeDepth();
    }; // class SymbolTable

    /* Exceptions thrown by some SymbolTable operations */
    class EmptySTException : public std::exception {
        public:
            const char * what() const noexcept override;
    };

    class DuplicateException : public std::exception {
        private:
            std::string message;
        public:
            DuplicateException(std::string symbol_name);
            const char * what() const noexcept override;
    };

    class FailedLookupException : public std::exception {
        private:
            std::string message;
        public:
            FailedLookupException(std::string symbol_name);
            const char * what() const noexcept override;
    };

    // Need to include definition in header for templates to work
    #include "symbol_table/SymbolTable.cpp"
}
