#include <limits.h>
#include "ast/context.hpp"

using namespace quant_c_ast;
using namespace type_info;

QuantumValue::QuantumValue() {
    qubitAddress = std::make_pair(-1, -1);
    classicalType = Type();
    isArg = false;
    isClassical = true;
    isShared = false;
    sharedName = "";
    qubitAddressSequence = std::vector<std::string>();
    isIntLiteral = false;
    intVal = 0;
    isSigned = false;

    isNull = true;
}

QuantumValue::QuantumValue(std::string name,
             std::pair<int, int> qubitAddress,
             bool isArg) : Symb(name) {
    this->qubitAddress = qubitAddress;
    classicalType = Type();
    this->isArg = isArg;
    isClassical = false;
    isShared = false;
    sharedName = "";
    isNull = false;
    isIntLiteral = false;
    intVal = 0;
    isSigned = false;

    // Generate address sequence
    qubitAddressSequence = std::vector<std::string>();
    for (int i = qubitAddress.first; i <= qubitAddress.second; i++) {
        std::string qubitAddr = (isArg) ? name + std::to_string(i) :
                                          std::to_string(i);
        qubitAddressSequence.push_back(qubitAddr);
    }
}

QuantumValue::QuantumValue(QuantumValue& lhs, QuantumValue& rhs) : Symb("") {
    qubitAddress = std::make_pair(-1, -1);
    classicalType = Type();
    isArg = false;
    isClassical = false;
    isShared = false;
    sharedName = "";
    isNull = false;
    isIntLiteral = false;
    intVal = 0;
    isSigned = false;

    // Generate address sequence by combining the two
    qubitAddressSequence = std::vector<std::string>(lhs.qubitAddressSequence);
    qubitAddressSequence.insert(qubitAddressSequence.end(),
                                rhs.qubitAddressSequence.begin(),
                                rhs.qubitAddressSequence.end());
}

QuantumValue::QuantumValue(std::string name,
             type_info::Type classicalType,
             bool isArg, std::string sharedName) : Symb(name) {
    this->qubitAddress = std::make_pair(-1, -1);
    this->classicalType = classicalType;
    this->isArg = isArg;
    isClassical = true;
    isShared = (sharedName.compare("") == 0) ? false : true;
    this->sharedName = sharedName;
    isNull = false;
    qubitAddressSequence = std::vector<std::string>();
    isIntLiteral = false;
    intVal = 0;
    isSigned = false;
}

QuantumValue::QuantumValue(std::intmax_t intval, bool isSigned) : Symb("") {
    qubitAddress = std::make_pair(-1, -1);
    classicalType = Type();
    isArg = false;
    isClassical = true;
    isShared = false;
    sharedName = "";
    isNull = false;

    isIntLiteral = true;
    this->intVal = intval;
    this->isSigned = isSigned;
    qubitAddressSequence = std::vector<std::string>();
}

// end QuantumValue



QuantumOpMod::QuantumOpMod(OpMod modifier, QuantumValue value) {
    this->modifier = modifier;
    this->value = value;
}

// end QuantumOpMod



QuantumContext::QuantumContext(int startQubit) {
    functionName = "";
    functionArgs = std::map<std::string, QuantumValue>();
    returnType = Type(TypeValue::VOID);
    declaredClassicalMemory = std::map<std::string, QuantumValue>();
    std::vector<QuantumOpMod> nextMods = std::vector<QuantumOpMod>();
    instructions = std::vector<std::string>();
    startingQubit = startQubit;
    nextUnassignedQubit = startingQubit;
    qubitAddresses = symbol_table::SymbolTable<QuantumValue>();
    qubitAddresses.openScope();
}

void QuantumContext::assignQubitAddress(std::string name,
                                        type_info::Type type,
                                        bool isArg) {

    int numQubits = Type::getTypeBitSize(type);

    std::pair<int, int> addressRange = std::make_pair(nextUnassignedQubit,
                                      nextUnassignedQubit + numQubits - 1);
    nextUnassignedQubit += numQubits;

    auto quantVal = QuantumValue(name, addressRange, isArg);
    qubitAddresses.insert(quantVal);

}

void QuantumContext::declareClassicalMemory(std::string name, type_info::Type type, bool isArg, std::string sharedName) {
    auto quantVal = QuantumValue(name, type, isArg, sharedName);
    declaredClassicalMemory.insert(std::make_pair(name, quantVal));
}

QuantumValue QuantumContext::getQuantumValue(std::string name) {
    try {
        return qubitAddresses.globalLookup(name);
    }
    catch (...) {}
    try {
        return functionArgs.at(name);
    }
    catch (...) {}
    try {
        return declaredClassicalMemory.at(name);
    }
    catch (...) {}

    return QUANTUM_NULL_VAL; // Error
}

std::string QuantumContext::genQUIL() {
    // Declare classical memory
    std::vector<std::string> declaredFuncMemory;
    std::vector<std::pair<std::string, int>> initSizes;
    for (auto& [name, qval]: declaredClassicalMemory) {
        if (qval.isClassical && qval.get_name() != "") { // Should always be classical
            std::string declareLine = "DECLARE " + name + " ";
            auto [type, size] = qval.classicalType.quantum_codegen_type();
            declareLine += type + "[" + std::to_string(size) + "]";

            // If declared memory is shared, their no need to initialize later
            if (qval.isShared) {
                declareLine += " SHARING " + qval.sharedName;
            }
            else {
                initSizes.push_back(std::make_pair(name, size));
            }

            declaredFuncMemory.push_back(declareLine);
        }
    }

    // Build DEFCIRCUIT definition
    std::string defcircuitName = "DEFCIRCUIT " + functionName;
    for (auto const& [name, qval] : functionArgs) {
        if (qval.isClassical) {
            defcircuitName += " " + name;
        }
        else {
            /*
            int startAddress = qval.qubitAddress.first;
            int endAddress = qval.qubitAddress.second;
            for (int i = startAddress; i <= endAddress; i++) {
                defcircuitName += " " + name + std::to_string(i);
            }
            */

            for (auto& qubit: qval.qubitAddressSequence) {
                defcircuitName += " " + qubit;
            }
        }
    }
    defcircuitName += " :";

    // Combine declarations, circuit definition, and instructions into final string
    std::string finalQuilCode = "";
    for (std::string& decls : declaredFuncMemory) {
        finalQuilCode += decls + "\n";
    }
    finalQuilCode += "\n" + defcircuitName + "\n" ;

    // Insert code to clear memory around instructions before
    // generating the final body
    for (auto const& [name, size]: initSizes) {
        std::string initStartInst = "MOVE " + name;
        std::string initInst = "";
        if (size == 1) {
            initInst = initStartInst + " 0; ";
        }
        else {
            for (int i = 0; i < size; i++) {
                initInst += initStartInst + "[" + std::to_string(i) + "] 0; ";
            }
        }
        instructions.insert(instructions.begin(), initInst);
    }
    for (int i = startingQubit; i < nextUnassignedQubit; i++) {
        instructions.push_back("RESET " + std::to_string(i) + ";");
    }

    // Generate body of circuit
    for (std::string& instLines : instructions) {
        finalQuilCode += "\t" + instLines + "\n";
    }
    finalQuilCode += "\n";

    return finalQuilCode;
}

void QuantumContext::buildMove(QuantumValue memoryVal, QuantumValue storeVal) {
    std::string moveInst;
    if (memoryVal.isClassical && storeVal.isClassical) {
        if (storeVal.isIntLiteral) {
            moveInst = "MOVE " + memoryVal.get_name() + " " +
                                 std::to_string(storeVal.intVal) + ";";
        }
        else if (memoryVal.classicalType != storeVal.classicalType) {
            auto memoryIsBits = memoryVal.classicalType.get_base_type_value() ==
                                                        TypeValue::BOOLEAN;
            auto storeIsBits = storeVal.classicalType.get_base_type_value() ==
                                                        TypeValue::BOOLEAN;
            auto memoryBitVec = (memoryIsBits) ? memoryVal :
                            getQuantumValue(memoryVal.get_name() + "_bits");
            auto storeBitVec = (storeIsBits) ? storeVal :
                                getQuantumValue(storeVal.get_name() + "_bits");
            if (memoryBitVec.isNull) {
                auto memoryName = memoryVal.get_name() + "_bits";
                int bitVecSize = Type::getTypeBitSize(memoryVal.classicalType);
                auto vecElementType = std::make_shared<Type>(TypeValue::BOOLEAN);
                Type bitVecType = Type(vecElementType, bitVecSize);
                this->declareClassicalMemory(memoryName, bitVecType, false,
                                             memoryVal.get_name());
                memoryBitVec = getQuantumValue(memoryName);
            }
            if (storeBitVec.isNull) {
                auto storeName = storeVal.get_name() + "_bits";
                int bitVecSize = Type::getTypeBitSize(storeVal.classicalType);
                auto vecElementType = std::make_shared<Type>(TypeValue::BOOLEAN);
                Type bitVecType = Type(vecElementType, bitVecSize);
                this->declareClassicalMemory(storeName, bitVecType, false,
                                             storeVal.get_name());
                storeBitVec = getQuantumValue(storeName);
            }

            int memorybitVecSize = Type::getTypeBitSize(memoryVal.classicalType);
            int storebitVecSize = Type::getTypeBitSize(storeVal.classicalType);

            for (int i = 0;
                 (i < memorybitVecSize) &&
                 (i < storebitVecSize); i++) {
                     moveInst += "MOVE " + memoryBitVec.get_name() +
                                    "[" + std::to_string(i) + "] " +
                                          storeBitVec.get_name() +
                                    "[" + std::to_string(i) + "]; ";
                 }
        }
        else {
            moveInst = "MOVE " + memoryVal.get_name() + " " +
                                 storeVal.get_name() + ";";
        }
    }

    instructions.push_back(moveInst);
}

void QuantumContext::buildQuantumInit(QuantumValue memoryVal, QuantumValue storeVal) {

    auto [type, size] = memoryVal.classicalType.quantum_codegen_type();
    // Add bit-vector as shared register if needed and not already declared

    std::string storeName;
    if (storeVal.isIntLiteral) {
        // If zero-initialized, no need to generate anything
        if (storeVal.intVal == 0) return;

        // Generate register to store int lit
        auto litVal = generateTempMemory(Type(TypeValue::INTEGER));
        instructions.push_back("MOVE " + litVal.get_name() + " " + std::to_string((unsigned int)storeVal.intVal) + ";");

        // Create a shared register to access the memory as bits
        storeName = litVal.get_name() + "_bits";
        int bitVecSize = Type::getTypeBitSize(Type(TypeValue::INTEGER));
        auto vecElementType = std::make_shared<Type>(TypeValue::BOOLEAN);
        Type bitVecType = Type(vecElementType, bitVecSize);
        this->declareClassicalMemory(storeName, bitVecType, false,
                                     litVal.get_name());
    }
    else if (storeVal.isClassical && type != "BIT" &&
            declaredClassicalMemory.find(storeVal.get_name() + "_bits") == declaredClassicalMemory.end()) {

        // Create a shared register to access the memory as bits
        storeName = storeVal.get_name() + "_bits";
        int bitVecSize = Type::getTypeBitSize(Type(TypeValue::INTEGER));
        auto vecElementType = std::make_shared<Type>(TypeValue::BOOLEAN);
        Type bitVecType = Type(vecElementType, bitVecSize);
        this->declareClassicalMemory(storeName, bitVecType, false,
                                     storeVal.get_name());
    }
    else if (storeVal.isClassical && type == "BIT") {
        storeName = storeVal.get_name();
    }
    else {
        storeName = storeVal.get_name() + "_bits";
    }

    // Generate quantum move instructions for initializer
    std::string moveInst;
    int classicalIndex = 0;
    /*
    for (int i = memoryVal.qubitAddress.first;
             i <= memoryVal.qubitAddress.second; i++) {
        moveInst += "QMOVE " + std::to_string(i) + " " +
                    storeName + "[" + std::to_string(classicalIndex) + "];";
        classicalIndex++;
    }
    */

    for (auto& qubit: memoryVal.qubitAddressSequence) {
        moveInst += "QMOVE " + qubit + " " +
                    storeName + "[" + std::to_string(classicalIndex) + "];";
        classicalIndex++;
    }
    instructions.push_back(moveInst);
}

QuantumValue QuantumContext::buildCircuitCall(std::string name, std::vector<QuantumValue> args) {
    std::string circuitCall = name + " ";
    for (auto& argVal: args) {
        if (argVal.isClassical) {
            circuitCall += argVal.get_name() + " ";
        }
        else {
            /*
            for (int i = argVal.qubitAddress.first; i <= argVal.qubitAddress.firsrt; i++) {
                std::string qubitName = std::to_string(i);
                if(argVal.isArg) {
                    qubitName = argVal.get_name() + qubitName;
                }
                circuitCall += qubitName + " ";
            }
            */

            for (auto& qubit: argVal.qubitAddressSequence) {
                circuitCall += qubit + " ";
            }
        }
    }
    circuitCall += ";";
    instructions.push_back(circuitCall);

    // TODO: figure out how to handle calls for quantum functions with return types
    return QUANTUM_NULL_VAL;
}

QuantumValue QuantumContext::buildMeasurement(QuantumValue qubits) {
    // Generate register to store measurement result
    int qubitsSize = qubits.qubitAddressSequence.size();
    auto bitVecElement = std::make_shared<Type>(TypeValue::BOOLEAN);
    Type measureRegisterType = Type(bitVecElement, qubitsSize);
    auto measureRegister = generateTempMemory(measureRegisterType);

    // Geenerate measurement instruction
    std::string measureInst;
    int registerIndex = 0;

    for (auto& qubit: qubits.qubitAddressSequence) {
        measureInst += "MEASURE " + qubit + " " +
                       measureRegister.get_name() + "[" +
                       std::to_string(registerIndex) + "]; ";
        registerIndex++;
    }
    instructions.push_back(measureInst);

    return measureRegister;
}


QuantumValue QuantumContext::buildUnaryGateOp(QuantumOpCode op,
                                      QuantumValue operand) {
    std::string opName = "I";
    switch(op) {
        case QuantumOpCode::I : break; // Already correct name
        case QuantumOpCode::X : opName = "X"; break;
        case QuantumOpCode::Y : opName = "Y"; break;
        case QuantumOpCode::Z : opName = "Z"; break;
        case QuantumOpCode::H : opName = "H"; break;
        case QuantumOpCode::S : opName = "S"; break;
        case QuantumOpCode::T : opName = "T"; break;
        default: break; // Assume identity operation for safety
    }

    // Check for modifiers and adjust op name accordingly
    for (auto& opMod : nextMods) {
        switch(opMod.modifier) {
            case OpMod::INVERSE: opName = "DAGGER " + opName; break;
            case OpMod::CONTROL: {
                /*
                for (int i = opMod.value.qubitAddress.first;
                     i <= opMod.value.qubitAddress.second; i++) {
                    std::string controlQubit = std::to_string(i);
                    if(opMod.value.isArg) {
                        controlQubit = opMod.value.get_name() + controlQubit;
                    }
                    opName = "CONTROLLED " + opName + " " + controlQubit;
                }
                */
                for (auto& qubit: opMod.value.qubitAddressSequence) {
                    opName = "CONTROLLED " + opName + " " + qubit;
                }
            } break;
        }
    }

    // Build the string to perform the gate on all of the qubits
    std::string opString = "";
    /*
    for (int i = operand.qubitAddress.first;
             i <= operand.qubitAddress.second; i++) {

        // If operand is local, just use the index
        // If the operand is an argument, append the name to the index
        std::string qubitName = std::to_string(i);
        if (operand.isArg) {
            qubitName = operand.get_name() + qubitName;
        }

        opString += (opName + " " + qubitName + "; ");
    }
    */

    for (auto& qubit: operand.qubitAddressSequence) {
        opString += (opName + " " + qubit + "; ");
    }

    // Insert the op string into the program code
    instructions.push_back(opString);

    // Clear out any previous modifiers
    nextMods.clear();

    return operand;
}

QuantumValue QuantumContext::buildBinaryGateOp(QuantumOpCode op,
                                       QuantumValue left,
                                       QuantumValue right) {
    std::string opName = "I2";
    switch(op) {
        case QuantumOpCode::CNOT :  opName = "CNOT"; break;
        case QuantumOpCode::SWAP :  opName = "SWAP"; break;
        case QuantumOpCode::ISWAP : opName = "ISWAP"; break;
        case QuantumOpCode::PSWAP : opName = "PSWAP"; break;
        default: break; // Assume do-nothing for safety
    }

    // Check for modifiers and adjust op name accordingly
    for (auto& opMod : nextMods) {
        switch(opMod.modifier) {
            case OpMod::INVERSE: opName = "DAGGER " + opName; break;
            case OpMod::CONTROL: {
                /*
                for (int i = opMod.value.qubitAddress.first;
                     i <= opMod.value.qubitAddress.second; i++) {
                    std::string controlQubit = std::to_string(i);
                    if(opMod.value.isArg) {
                        controlQubit = opMod.value.get_name() + controlQubit;
                    }
                    opName = "CONTROLLED " + opName + " " + controlQubit;
                }
                */
                for (auto& qubit: opMod.value.qubitAddressSequence) {
                    opName = "CONTROLLED " + opName + " " + qubit;
                }
            } break;
        }
    }

    // Build the string to perform the gate on all of  the qubits
    std::string opString = "";
    /*
    for (int i = left.qubitAddress.first, j = right.qubitAddress.first;
             i <= left.qubitAddress.second || j < right.qubitAddress.second; i++, j++) {

        std::string LHSQubitName = std::to_string(i);
        std::string RHSQubitName = std::to_string(j);
        if (left.isArg) {
            LHSQubitName = left.get_name() + LHSQubitName;
        }
        if (right.isArg) {
            RHSQubitName = right.get_name() + RHSQubitName;
        }

        opString += (opName + " " + LHSQubitName +
                              " " + RHSQubitName + "; ");
    }
    */
    for (auto itLeft = left.qubitAddressSequence.begin(),
              itRight = right.qubitAddressSequence.begin();
              itLeft != left.qubitAddressSequence.end() &&
              itRight != right.qubitAddressSequence.end();
              itLeft++, itRight++) {
        opString += (opName + " " + *itLeft +
                              " " + *itRight + "; ");
    }
    opString += ";";

    // Insert the op string into the program code
    instructions.push_back(opString);

    // Clear out any previous modifiers
    nextMods.clear();

    return QuantumValue(left, right);
}

void QuantumContext::buildReturn(QuantumValue returnVal) {
    std::string moveInst;
    std::string funcReturnName = functionName + "_ro";
    std::string retValName = returnVal.get_name();
    if (returnVal.classicalType.typeValCompare(returnType) == false) {
        auto returnRegIsBits = returnType.get_base_type_value() ==
                                                    TypeValue::BOOLEAN;
        auto returnRegBitVec = (returnRegIsBits) ?
                                    getQuantumValue(funcReturnName) :
                                    getQuantumValue(funcReturnName + "_bits");
        bool returnValIsBits = returnVal.classicalType.get_base_type_value() ==
                               TypeValue::BOOLEAN;
        auto returnValBitVec = (returnValIsBits) ? returnVal :
                               getQuantumValue(returnVal.get_name() + "_bits");
        if (returnValBitVec.isNull) {
            auto returnValName = returnVal.get_name() + "_bits";
            int bitVecSize = Type::getTypeBitSize(returnVal.classicalType);
            auto vecElementType = std::make_shared<Type>(TypeValue::BOOLEAN);
            Type bitVecType = Type(vecElementType, bitVecSize);
            this->declareClassicalMemory(returnValName, bitVecType, false,
                                         returnVal.get_name());
            returnValBitVec = getQuantumValue(returnValName);
        }

        int returnValSize = Type::getTypeBitSize(returnRegBitVec.classicalType);
        int returnRegSize = Type::getTypeBitSize(returnValBitVec.classicalType);

        for (int i = 0; (i < returnValSize) && (i < returnRegSize); i++) {
            moveInst += "MOVE " +
                        returnRegBitVec.get_name() + "[" + std::to_string(i) + "] " +
                        returnValBitVec.get_name() + "[" + std::to_string(i) + "]; ";
        }
    } else {
        moveInst = "MOVE " + funcReturnName + " " + retValName + ";";
    }

    instructions.push_back(moveInst);
}

void QuantumContext::setFunctionName(std::string name) {
    functionName = name;
}

void QuantumContext::addFunctionArgument(std::string name, type_info::Type type) {
    QuantumValue argVal;
    if (type.get_base_type_storage() == StorageSpecVal::QUANTUM) {
        int numQubits = Type::getTypeBitSize(type);
        std::pair<int, int> argIndices = std::make_pair(0, numQubits-1);
        argVal = QuantumValue(name, argIndices, true);
    }
    else {
        argVal = QuantumValue(name, type);
    }

    functionArgs.insert(std::make_pair(name, argVal));
}

void QuantumContext::setFunctionReturnType(type_info::Type type) {
    returnType = type;
    returnType.storage = StorageSpecVal::AUTO;

    // Build register for return type if needed
    // Also build bit vector overlapping it if needed
    if (type.get_base_type_value() != TypeValue::VOID &&
        type.get_base_type_value() != TypeValue::UNKNOWN &&
        type.get_base_type_value() != TypeValue::ERROR) {
            this->declareClassicalMemory(functionName + "_ro", type);

            if (type.get_base_type_value() != TypeValue::BOOLEAN) {
                int bitVecSize = Type::getTypeBitSize(type);
                auto bitVecElementType = std::make_shared<Type>(TypeValue::BOOLEAN);
                Type bitVecType = Type(bitVecElementType, bitVecSize);
                this->declareClassicalMemory(functionName + "_ro_bits", bitVecType, false, functionName + "_ro");
            }

    }

}

void QuantumContext::insertOpMod(QuantumOpMod mod) {
    nextMods.push_back(mod);
}

QuantumValue QuantumContext::generateTempMemory(type_info::Type type) {
    // Generate temporary main
    std::string memName = functionName + "_tmp";
    if (tmpId) memName += std::to_string(tmpId);
    tmpId++;

    // Save memory alue to generate later and return this value
    auto classicalMemVal = QuantumValue(memName, type);
    declaredClassicalMemory.insert(std::make_pair(memName, classicalMemVal));
    return classicalMemVal;
}

// end QuantumContext
