#include <iostream>
#include <iomanip>
#include <algorithm>
#include <iterator>
#include <string_view>
#include <cmath>
#include <exception>
#include <stdexcept>
#include "symbol_table/symbol_table.hpp"
#include "ast/ast.hpp"
#include "quant_c_parse.hpp"

using namespace quant_c_ast;
using namespace symbol_table;
using namespace type_info;
using namespace quant_c_compiler;

std::shared_ptr<NullTypeNode> TypeNode::NULL_VAL = std::make_shared<NullTypeNode>();

TypeNode::TypeNode(ASTNodeKind nodeKind) : ASTNode(nodeKind) {
    type = Type(TypeValue::UNKNOWN);
}
TypeNode::TypeNode(ASTNodeKind nodeKind, int line, int col, Type type ) : ASTNode(nodeKind, line, col) {
    this->type = type;
}
bool TypeNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_TypeNode &&
           nodeKind < ASTNodeKind::AST_ArgDeclNode;
} // classof
// class TypeNode

NullTypeNode::NullTypeNode() : TypeNode(ASTNodeKind::AST_NullTypeNode) {}
bool NullTypeNode::isNull() {
    return true;
}
void NullTypeNode::unparse(int indent) { /* no action */ }
bool NullTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_NullTypeNode;
} // classof
// class NullTypeNode


std::shared_ptr<NullTypeSpecNode> TypeSpecNode::NULL_VAL = std::make_shared<NullTypeSpecNode>();

TypeSpecNode::TypeSpecNode(ASTNodeKind nodeKind) : TypeNode::TypeNode(nodeKind) {}
TypeSpecNode::TypeSpecNode(ASTNodeKind nodeKind, int line, int col, type_info::Type type) : TypeNode::TypeNode(nodeKind, line, col, type) {}
bool TypeSpecNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_TypeSpecNode &&
           nodeKind <= ASTNodeKind::AST_CustomTypeNode;
} // classof
// class TypeSpecNode

NullTypeSpecNode::NullTypeSpecNode() : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_NullTypeSpecNode) {}
bool NullTypeSpecNode::isNull() { return true; }
void NullTypeSpecNode::unparse(int indent) {}
bool NullTypeSpecNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_NullTypeSpecNode;
} // classof
// class NullTypeSpecNode

IntTypeNode::IntTypeNode(int line, int col) :
TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_NullTypeSpecNode, line, col, Type(TypeValue::INTEGER)) {}

void IntTypeNode::unparse(int indent) {
    std::cout << "int";
}

bool IntTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_IntTypeNode;
} // classof
// class IntTypeNode

BoolTypeNode::BoolTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_BoolTypeNode, line, col, Type(TypeValue::BOOLEAN)) {}

void BoolTypeNode::unparse(int indent) {
    std::cout << "bool";
}
bool BoolTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_BoolTypeNode;
} // classof
// class BoolTypeNode


CharTypeNode::CharTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_CharTypeNode, line, col, Type(TypeValue::CHARACTER)) {}
void CharTypeNode::unparse(int indent) {
    std::cout << "char";
}
bool CharTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_CharTypeNode;
} // classof
// class CharTypeNode


VoidTypeNode::VoidTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_VoidTypeNode, line, col, Type(TypeValue::VOID)) {}

void VoidTypeNode::unparse(int indent) {
    std::cout << "void";
}
bool VoidTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_VoidTypeNode;
} // classof
// class VoidTypeNode

ShortTypeNode::ShortTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_ShortTypeNode, line, col, Type(TypeValue::SHORT)) {}

void ShortTypeNode::unparse(int indent) {
    std::cout << "short";
}
bool ShortTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ShortTypeNode;
} // classof
// class ShortTypeNode


LongTypeNode::LongTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_LongTypeNode, line, col, Type(TypeValue::LONG)) {}

void LongTypeNode::unparse(int indent) {
    std::cout << "long";
}
bool LongTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_LongTypeNode;
} // classof
// class LongTypeNode


FloatTypeNode::FloatTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_FloatTypeNode, line, col, Type(TypeValue::FLOAT)) {}

void FloatTypeNode::unparse(int indent) {
    std::cout << "float";
}
bool FloatTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_FloatTypeNode;
} // classof
// class FloatTypeNode


DoubleTypeNode::DoubleTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_DoubleTypeNode, line, col, Type(TypeValue::DOUBLE)) {}

void DoubleTypeNode::unparse(int indent) {
    std::cout << "double";
}
bool DoubleTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_DoubleTypeNode;
} // classof
// class DoubleTypeNode


SignedTypeNode::SignedTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_SignedTypeNode, line, col, Type(TypeValue::VOID, true)) {}

void SignedTypeNode::unparse(int indent) {
    std::cout << "signed";
}
bool SignedTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_SignedTypeNode;
} // classof
// class SignedTypeNode


UnsignedTypeNode::UnsignedTypeNode(int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_TypeSpecNode, line, col, Type(TypeValue::VOID, false)) {}

void UnsignedTypeNode::unparse(int indent) {
    std::cout << "unsigned";
}
bool UnsignedTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_UnsignedTypeNode;
} // classof
// class UnsignedTypeNode


CustomTypeNode::CustomTypeNode(std::string type_name, int line, int col) : TypeSpecNode::TypeSpecNode(ASTNodeKind::AST_CustomTypeNode, line, col, Type(TypeValue::UNKNOWN)) {
    this->type_name = type_name;
}

void CustomTypeNode::unparse(int indent) {
    std::cout << type_name;
}
bool CustomTypeNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_CustomTypeNode;
} // classof
// class CustomTypeNode


StorageClassSpecNode::StorageClassSpecNode(ASTNodeKind nodeKind, int line, int col) : TypeNode::TypeNode(nodeKind, line, col, Type(TypeValue::VOID)) {}
bool StorageClassSpecNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_StorageClassSpecNode &&
           nodeKind <= ASTNodeKind::AST_QuantumNode;
} // classof
// class StorageClassSpecNode


TypeDefNode::TypeDefNode(int line, int col) : StorageClassSpecNode::StorageClassSpecNode(ASTNodeKind::AST_TypeDefNode, line, col) {}

void TypeDefNode::unparse(int indent) {
    std::cout << "typedef";
}
bool TypeDefNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_TypeDefNode;
} // classof
// class TypeDefNode


ExternNode::ExternNode(int line, int col) : StorageClassSpecNode::StorageClassSpecNode(ASTNodeKind::AST_ExternNode, line, col) {}

void ExternNode::unparse(int indent) {
    std::cout << "extern";
}
bool ExternNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ExternNode;
} // classof
// class ExternNode


StaticNode::StaticNode(int line, int col) : StorageClassSpecNode::StorageClassSpecNode(ASTNodeKind::AST_StaticNode, line, col) {}

void StaticNode::unparse(int indent) {
    std::cout << "static";
}
bool StaticNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_StaticNode;
} // classof
// class StaticNode


AutoNode::AutoNode(int line, int col) : StorageClassSpecNode::StorageClassSpecNode(ASTNodeKind::AST_AutoNode, line, col) {}

void AutoNode::unparse(int indent) {
    std::cout << "auto";
}
bool AutoNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_AutoNode;
} // classof
// class AutoNode


RegisterNode::RegisterNode(int line, int col) : StorageClassSpecNode::StorageClassSpecNode(ASTNodeKind::AST_RegisterNode, line, col) {}

void RegisterNode::unparse(int indent) {
    std::cout << "register";
}
bool RegisterNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_RegisterNode;
} // classof
// class RegisterNode


QuantumNode::QuantumNode(int line, int col) : StorageClassSpecNode::StorageClassSpecNode(ASTNodeKind::AST_QuantumNode, line, col) {}

void QuantumNode::unparse(int indent) {
    std::cout << "quantum";
}
bool QuantumNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_QuantumNode;
} // classof
// class QuantumNode

TypeQualifierNode::TypeQualifierNode(ASTNodeKind nodeKind, int line, int col) : TypeNode::TypeNode(nodeKind, line, col, Type(TypeValue::VOID)) {}
bool TypeQualifierNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_TypeQualifierNode &&
           nodeKind <= ASTNodeKind::AST_VolatileNode;
} // classof
// class TypeQualifierNode


ConstNode::ConstNode(int line, int col) : TypeQualifierNode::TypeQualifierNode(ASTNodeKind::AST_ConstNode, line, col) {}

void ConstNode::unparse(int indent) {
    std::cout << "const";
}
bool ConstNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ConstNode;
} // classof
// class ConstNode


VolatileNode::VolatileNode(int line, int col) : TypeQualifierNode::TypeQualifierNode(ASTNodeKind::AST_VolatileNode, line, col) {}

void VolatileNode::unparse(int indent) {
    std::cout << "volatile";
}
bool VolatileNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_VolatileNode;
} // classof
// class VolatileNode


PointerNode::PointerNode(PointerNode::TypeQualifierList qualifiers,
                         int line, int col)
: TypeNode::TypeNode(ASTNodeKind::AST_ConstNode, line, col, Type(TypeValue::VOID)) {
    this->qualifiers = qualifiers;
}

void PointerNode::unparse(int indent) {
    std::cout << "* ";
    for (auto qualities : qualifiers) {
        qualities->unparse();
    }
}
bool PointerNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_PointerNode;
} // classof
// class PointerNode


TypeNameNode::TypeNameNode(SpecQualList spec_qual_list,
        std::shared_ptr<AbstractDeclaratorNode> abstract_declarator,
        int line, int col) : TypeNode::TypeNode(ASTNodeKind::AST_TypeNameNode, line, col, Type(TypeValue::VOID)) {
            this->spec_qual_list = spec_qual_list;
            this->abstract_declarator = abstract_declarator;
}

void TypeNameNode::unparse(int indent) {
    for (auto& specQual : spec_qual_list) {
        specQual->unparse();
    }
    abstract_declarator->unparse();
}

type_info::Type TypeNameNode::generateType() {
    Type typeSpecs = Type();
    typeSpecs.isScalar = true;
    bool set_signedness = false;
    bool set_quality = false;

    // Loop through specifiers and qualifiers to build basic type
    // ex. int, volatile int, const char
    for (auto& spec : spec_qual_list) {
        auto type_node = spec.get();
        if (llvm::isa<TypeSpecNode>(type_node)) {
            if (llvm::isa<LongTypeNode>(type_node)) {
                if (typeSpecs.value == TypeValue::UNKNOWN ||
                    typeSpecs.value == TypeValue::INTEGER) {
                    typeSpecs.value = TypeValue::LONG;
                }
                else if (typeSpecs.value == TypeValue::LONG) {
                    typeSpecs = TypeValue::LONGLONG;
                }
                else {
                    mustBe(false, error() + "both 'long' and '" +
                           typeSpecs.get_value_string() +
                           "' in declaration specifiers");
                    return Type(TypeValue::ERROR);
                }
            }
            else if (llvm::isa<ShortTypeNode>(type_node)) {
                if (typeSpecs.value == TypeValue::UNKNOWN ||
                    typeSpecs.value == TypeValue::INTEGER) {
                    typeSpecs.value = TypeValue::SHORT;
                }
                else if (typeSpecs.value == TypeValue::SHORT) {
                    mustBe(false, error() + "duplicate 'short'");
                    return Type(TypeValue::ERROR);
                }
                else {
                    mustBe(false, error() + "both 'short' and '" +
                           typeSpecs.get_value_string() +
                           "' in declaration specifiers");
                    return Type(TypeValue::ERROR);
                }
            }
            else if (llvm::isa<SignedTypeNode>(type_node)) {
                if (set_signedness) {
                    if (typeSpecs.isSigned) {
                        mustBe(false, error() + "duplicate 'signed'");
                    }
                    else {
                        mustBe(false, error() + "both 'signed' and 'unsigned' in declaration specifiers");
                    }
                    return Type(TypeValue::ERROR);
                } else {
                    typeSpecs.isSigned = true;
                    set_signedness = true;
                }
            }
            else if (llvm::isa<UnsignedTypeNode>(type_node)) {
                if (set_signedness) {
                    if (!typeSpecs.isSigned) {
                        mustBe(false, error() + "duplicate 'unsigned'");
                    }
                    else {
                        mustBe(false, error() + "both 'signed' and 'unsigned' in declaration specifiers");
                    }
                    return Type(TypeValue::ERROR);
                } else {
                    typeSpecs.isSigned = false;
                    set_signedness = true;
                }
            }
            else if (typeSpecs.value == TypeValue::UNKNOWN) {
                typeSpecs.value = spec->type.value;
            }
            else {
                mustBe(false, error() +
                       "two or more data types in declaration specifiers");
                return Type(TypeValue::ERROR);
            }
        }
        else if (llvm::isa<StorageClassSpecNode>(type_node)) {
            mustBe(false, error() +
                    "storage class specifier for type name");
        }
        else if (llvm::isa<TypeQualifierNode>(type_node)) {
            set_quality = true;
            if (typeSpecs.quality == TypeQualityVal::VOLATILE_CONST) {
                continue; // Any additional qualifiers are ignored
            }
            else if (llvm::isa<VolatileNode>(type_node)) {
                typeSpecs.quality =
                    (typeSpecs.quality == TypeQualityVal::CONST) ?
                    TypeQualityVal::VOLATILE_CONST :
                    TypeQualityVal::VOLATILE;

            } else { // llvm::isa<ConstNode>(spec)
                typeSpecs.quality =
                    (typeSpecs.quality == TypeQualityVal::VOLATILE) ?
                    TypeQualityVal::VOLATILE_CONST :
                    TypeQualityVal::CONST;
            }
        }
        else {
            mustBe(false, error() + "Unimplemented type used");
            return Type(TypeValue::ERROR);
        }
    }

    // Establish defaults if left implicit in the type specs
    if (typeSpecs.value == TypeValue::UNKNOWN) {
        typeSpecs.value = TypeValue::INTEGER;
    }
    if (!set_signedness) {
        typeSpecs.isSigned = true;
    }
    typeSpecs.storage = StorageSpecVal::AUTO;
    if (!set_quality) {
        typeSpecs.quality = TypeQualityVal::NONE;
    }

    // Build more complex type info (pointer, array, etc.) and
    // set base type to basic type
    Type declType = abstract_declarator->generateAbstractType();
    declType.set_base_type(typeSpecs);

    return declType;
}

void TypeNameNode::checkSemantics() {
    type = generateType();
}

bool TypeNameNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_TypeNameNode;
}
// class TypeNameNode

// end TypeNode subclasses
