#include "token/token.hpp"
#include "quant_c_parse.hpp"
#include <regex>

using namespace quant_c_token;
using namespace quant_c_compiler;

/* Exceptions */
const char * InvalidTokenException::what() const noexcept {
    auto error_message =
        std::string("Invalid token found (line ") + std::to_string(linenum) +
        " column " + std::to_string(colnum) + "): " + bad_token;
    return error_message.c_str();
}

/* Token */
Token::Token(int line, int col) {
    this->linenum = line;
    this->colnum = col;
}

/* BoolLitToken */
BoolLitToken::BoolLitToken(bool boolValue, int line, int col):
    Token {line, col} {
        this->boolValue = boolValue;
}

bool BoolLitToken::get_value() {
    return this->boolValue;
}

/* CharLitToken */
CharLitToken::CharLitToken(char val, int line, int col):
    Token {line, col} {
        this->charValue = val;
}

char CharLitToken::get_value() {
    return this->charValue;
}

std::string CharLitToken::printCharValue() {
    switch(this->charValue) {
            case '\n': return "'\\n'";
            case '\t': return "'\\t'";
            case '\\': return "'\\\\'";
            case '\'': return "'\\\''";
            default: return "'" + std::string(1, charValue) + "'";
    }
}

/* IdentifierToken */
IdentifierToken::IdentifierToken(std::string name, int line, int col):
    Token {line, col} {
        this->name = name;
}

std::string IdentifierToken::get_value() {
    return this->name;
}

/* IntLitToken */
IntLitToken::IntLitToken(std::intmax_t val, IntLitSize size, bool isSigned, int line, int col) :
    Token {line, col} {
        this->intValue = val;
        this->size = size;
        this->isSigned = isSigned;
}

std::intmax_t IntLitToken::get_value() {
    return this->intValue;
}

IntLitSize IntLitToken::get_size() {
    return this->size;
}

bool IntLitToken::is_signed() {
    return this->isSigned;
}

/* FloatLitToken */
FloatLitToken::FloatLitToken(long double val, FloatLitSize size, int line, int col) : Token {line, col} {
        this->floatValue = val;
        this->size = size;
}

double FloatLitToken::get_value() {
    return this->floatValue;
}

FloatLitSize FloatLitToken::get_size() {
    return this->size;
}

/* ErrorToken */
ErrorToken::ErrorToken(std::string value, int line, int col):
    Token {line, col} {
        this->badToken = value;
}

std::string ErrorToken::get_value() {
    return this->badToken;
}

/* StringLitToken */
std::unordered_map<std::string, std::string> StringLitToken::_static_init() {
    auto static_map = std::unordered_map<std::string, std::string>();
    static_map.insert(std::make_pair("\\\"",""));
    static_map.insert(std::make_pair("\\\\\\\\","\\"));
    static_map.insert(std::make_pair("\\\\\\\"","\""));
    static_map.insert(std::make_pair("\\\\t","\t"));
    static_map.insert(std::make_pair("\\\\n","\n"));
    return static_map;
}

std::unordered_map<std::string, std::string> StringLitToken::substitutions = StringLitToken::_static_init();

StringLitToken::StringLitToken(std::string text, int line, int col):
    Token {line, col} {
        this->string = text;

        // Minimize string
        this->reducedString = text;
        for (const auto& pattern_sub_pair : this->substitutions) {
            this->reducedString = std::regex_replace(this->reducedString,
                std::regex(pattern_sub_pair.first), pattern_sub_pair.second);
        }
}

std::string StringLitToken::get_value(bool raw_string) {
    if (raw_string) {
        return this->string;
    } else {
        return this->reducedString;
    }
}

/* IdentifierToken */
TypeNameToken::TypeNameToken(std::string name, int line, int col):
    Token {line, col} {
        this->name = name;
}

std::string TypeNameToken::get_value() {
    return this->name;
}
